"""
   Copyright 2024 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

# Example of how to utilize the Folium-based viewer
####################################################################################################
import geeViz.getImagesLib as gil
import  geeViz.foliumView as fv
ee = fv.ee
Map = fv.foliumMapper()

Map.port = 1234
Map.clearMap()
####################################################################################################
# Bring in some S2 data
studyArea = ee.Geometry.Polygon(
        [[[-113.21807278537877, 41.786028237932015],
          [-113.21807278537877, 40.595571243156144],
          [-111.82280911350377, 40.595571243156144],
          [-111.82280911350377, 41.786028237932015]]], None, False)

# Get some example images to view
s2s = gil.getProcessedSentinel2Scenes(studyArea,2022,2023,120,150)
postComposite = s2s.filter(ee.Filter.calendarRange(2023,2023,'year')).median()
Map.addLayer(postComposite,gil.vizParamsFalse,'S2 Median 2023')
preComposite = s2s.filter(ee.Filter.calendarRange(2022,2022,'year')).median()
Map.addLayer(preComposite,gil.vizParamsFalse,'S2 Median 2022',False)

# Center on the study area and view it
Map.centerObject(studyArea)
Map.addLayer(studyArea,{'strokeColor':'F00','strokeWidth':5},'Study Area')
####################################################################################################
# View the map
Map.view()




"""
   Copyright 2024 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
#Script to allow GEE objects to be viewed in a web viewer
#Intended to work within the geeViz package
######################################################################
#Import modules
import ee,sys,os,webbrowser,json,socket,subprocess,site,time,requests,google
from google.auth.transport import requests as gReq
from google.oauth2 import service_account

from threading import Thread
from urllib.parse import urlparse
from IPython.display import IFrame,display, HTML
if sys.version_info[0] < 3:
    import SimpleHTTPServer, SocketServer
else:
    import http.server, socketserver 
creds_path = ee.oauth.get_credentials_path()
creds_dir = os.path.dirname(creds_path)
if not os.path.exists(creds_dir):os.makedirs(creds_dir)

IS_COLAB = ee.oauth._in_colab_shell()#"google.colab" in sys.modules
IS_WORKBENCH = os.getenv("DL_ANACONDA_HOME") != None
if IS_COLAB:
  from google.colab.output import eval_js

######################################################################
# Functions to handle various initialization/authentication workflows to try to get a user an initialized instance of ee

# Function to have user input a project id if one is still needed
def setProject(id):
    global project_id
    project_id = id
    ee.data.setCloudApiUserProject(project_id)
def getProject(overwrite=False):
    global project_id
    provided_project = '{}.proj_id'.format(creds_path)
    provided_project = os.path.normpath(provided_project)
    
    current_project = ee.data._cloud_api_user_project
    
    if (current_project == None and not os.path.exists(provided_project)) or overwrite:
                    
            
        project_id = input('Please enter GEE project ID: ')
        
        print('You entered: {}'.format(project_id))
        o = open(provided_project,'w')
        o.write(project_id)
        o.close()
    
    if current_project != None:
        project_id = current_project
    elif os.path.exists(provided_project):
        o = open(provided_project,'r')
        project_id=o.read()
        print('Cached project id file path: {}'.format(provided_project))
        print('Cached project id: {}'.format(project_id))
        o.close()
    ee.data.setCloudApiUserProject(project_id)
    
    return project_id
def verified_initialize(project=None):
    ee.Initialize(project=project)
    z = ee.Number(1).getInfo()
    print('Successfully initialized')
# Function to handle various exceptions to initializing to GEE

def robustInitializer():
    global project_id
   
    try:
        z = ee.Number(1).getInfo()
    except:
        print('Initializing GEE')
        if not ee.oauth._valid_credentials_exist():
            ee.Authenticate()
        try:
            verified_initialize(project=ee.data._cloud_api_user_project)
        except Exception as E:
            # print(E)
            if str(E).find('Reauthentication is needed') >- 1:
                ee.Authenticate(force=True)

            if str(E).find('no project found. Call with project') or str(E).find('project is not registered') > -1 or str(E).find(' quota project, which is not set by default')>-1:
                project_id=getProject()
                
            else:
                project_id = None
            try:
                verified_initialize(project=project_id)
            except Exception as E:
                print(E)
                try:
                    project_id=getProject(overwrite=True)
                    verified_initialize(project=project_id)
                except Exception as E:
                    print(E)
       
        ee.data.setCloudApiUserProject(project_id)
setProject(ee.data._cloud_api_user_project)
robustInitializer()
######################################################################
#Set up GEE and paths
# robustInitializer()
geeVizFolder = 'geeViz'
geeViewFolder = 'geeView'
#Set up template web viewer
#Do not change
cwd = os.getcwd()

paths = sys.path

# gee_py_modules_dir = site.getsitepackages()[-1]
# py_viz_dir = os.path.join(gee_py_modules_dir,geeVizFolder)
py_viz_dir = os.path.dirname(__file__)
# os.chdir(py_viz_dir)
print('geeViz package folder:', py_viz_dir)

#Specify location of files to run
template = os.path.join(py_viz_dir,geeViewFolder,'index.html')
ee_run_dir =  os.path.join(py_viz_dir, geeViewFolder,'src/gee/gee-run/')
if os.path.exists(ee_run_dir) == False:os.makedirs(ee_run_dir)


######################################################################
######################################################################
#Functions

######################################################################
# Function to check if being run inside a notebook
# Taken from: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def is_notebook():
    return ee.oauth._in_jupyter_shell()
    # try:
    #     shell = get_ipython().__class__.__name__
    #     if shell == 'ZMQInteractiveShell':
    #         return True   # Jupyter notebook or qtconsole
    #     elif shell == 'TerminalInteractiveShell':
    #         return False  # Terminal running IPython
    #     else:
    #         return False  # Other type (?)
    # except NameError:
    #     return False      # Probably standard Python interpreter
######################################################################
# Function for cleaning trailing .... in accessToken
def cleanAccessToken(accessToken):
    while accessToken[-1] == '.': accessToken = accessToken[:-1]
    return accessToken
# Function to get domain base without any folders
def baseDomain(url):
    url_parts = urlparse(url)
    return f'{url_parts.scheme}://{url_parts.netloc}'
# Function for using default GEE refresh token to get an access token for geeView
# Updated 12/23 to reflect updated auth methods for GEE
def refreshToken():
    credentials = ee.data.get_persistent_credentials()
    credentials.refresh(gReq.Request())
    accessToken = credentials.token
    # print(credentials.to_json())
    accessToken = cleanAccessToken(accessToken)
    return accessToken
    
# Function for using a GEE white-listed service account key to get an access token for geeView
def serviceAccountToken(service_key_file_path):
    try:
        credentials = service_account.Credentials.from_service_account_file(service_key_file_path, scopes=ee.oauth.SCOPES)
        credentials.refresh(gReq.Request())
        accessToken = credentials.token
        accessToken = cleanAccessToken(accessToken)
        return accessToken
    except Exception as e:
        print(e)
        print('Failed to utilize service account key file.')
        return None
#Function for running local web server
def run_local_server(port = 8001):
    if sys.version[0] == '2':
        server_name = 'SimpleHTTPServer'
    else:
        server_name = 'http.server'
    cwd = os.getcwd()
    os.chdir(py_viz_dir)
    # print('cwd',os.getcwd())
    python_path = sys.executable
    if python_path.find('pythonw')>-1:python_path = python_path.replace('pythonw','python')
    c = '"{}" -m {}  {}'.format(python_path, server_name,port)
    print('HTTP server command:',c)
    subprocess.Popen(c,shell = True)
    os.chdir(cwd)



#Function to see if port is active
def isPortActive(port = 8001):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(2)                                      #2 Second Timeout
    result = sock.connect_ex(('localhost',port))
    if result == 0:
        return True
    else:
        return False

#Set up map object
class mapper:
    def __init__(self,port = 8001):
        self.port = port
        self.layerNumber = 1
        self.idDictList = []
        self.mapCommandList  = []
        self.ee_run_name = 'runGeeViz'

        self.isNotebook = ee.oauth._in_jupyter_shell()
        self.isColab = ee.oauth._in_colab_shell()#"google.colab" in sys.modules

        self.proxy_url = None

        self.refreshTokenPath = ee.oauth.get_credentials_path()
        self.serviceKeyPath = None
        self.queryWindowMode = 'sidePane'
        self.project = project_id
    #Function for adding a layer to the map
    def addLayer(self,image,viz = {},name= None,visible= True):
        if name == None:
            name = 'Layer '+str(self.layerNumber)
            self.layerNumber+=1
        print('Adding layer: ' +name)

        # Handle reducer if ee object is given
        if 'reducer' in viz.keys():
            # if str(type(viz['reducer']))=="<class 'ee.Reducer'>":
            #     viz['reducer'] = eval(viz['reducer'])

            try:
                viz['reducer'] = viz['reducer'].serialize()
            except Exception as e:
                try:
                    viz['reducer'] = eval(viz['reducer']).serialize()
                except Exception as e:# Most likely it's already serialized
                    e = e
        #Get the id and populate dictionary
        idDict = {}
        if not isinstance(image, dict):
            image=image.serialize()
            idDict['item'] = image
            idDict['function'] = 'addSerializedLayer'
        # Handle passing in geojson vector layers
        else:
            idDict['item'] = image
            viz['layerType']='geoJSONVector'
            idDict['function'] = 'addLayer'

        idDict['name'] = name 
        idDict['visible'] = str(visible).lower()
        idDict['viz'] = json.dumps(viz, sort_keys=False)
        
        self.idDictList.append(idDict)

    #Function for adding a layer to the map
    def addTimeLapse(self,image,viz = {},name= None,visible= True):
        if name == None:
            name = 'Layer '+str(self.layerNumber)
            self.layerNumber+=1
        print('Adding layer: ' +name)

        # Handle reducer if ee object is given - delete it
        if 'reducer' in viz.keys():
            del viz['reducer']
        #Get the id and populate dictionary
        idDict = {}#image.getMapId()
        idDict['item'] = image.serialize()
        idDict['name'] = name 
        idDict['visible'] = str(visible).lower()
        idDict['viz'] = json.dumps(viz, sort_keys=False)
        idDict['function'] = 'addSerializedTimeLapse'
        self.idDictList.append(idDict)

    # Function for setting the map zoom
    def setZoom(self,zoom):
        self.mapCommandList.append(f'map.setZoom({zoom})')
    #Function for centering on a GEE object that has a geometry
    def centerObject(self,feature,zoom = None):
        try:
            bounds = json.dumps(feature.geometry().bounds(100,'EPSG:4326').getInfo())
        except Exception as e:
            bounds = json.dumps(feature.bounds(100,'EPSG:4326').getInfo())
        command = 'synchronousCenterObject('+bounds+')'
        
        self.mapCommandList.append(command)

        if zoom != None:
            self.setZoom(zoom)
    #Function for launching the web map after all adding to the map has been completed
    def view(self,open_browser = None, open_iframe = None,iframe_height = 525):
        print('Starting webmap')

        # Get access token
        if self.serviceKeyPath == None:
            print('Using default refresh token for geeView')
            self.accessToken = refreshToken()
        else:
            print('Using service account key for geeView:',self.serviceKeyPath)
            self.accessToken = serviceAccountToken(self.serviceKeyPath)
            if self.accessToken == None:
                print('Trying to authenticate to GEE using persistent refresh token.')
                self.accessToken = refreshToken(self.refreshTokenPath)
        #Set up js code to populate
        lines = "var layerLoadErrorMessages=[];showMessage('Loading',staticTemplates.loadingModal[mode]);\nfunction runGeeViz(){\n"


        #Iterate across each map layer to add js code to
        for idDict in self.idDictList:
            t ="Map.{}({},{},'{}',{});".format(idDict['function'],idDict['item'],idDict['viz'],idDict['name'],str(idDict['visible']).lower())
            t = 'try{\n\t'+t+'\n}catch(err){\n\tlayerLoadErrorMessages.push("Error loading: '+idDict['name']+'<br>GEE "+err);}\n'
            lines += t
        lines += 'if(layerLoadErrorMessages.length>0){showMessage("Map.addLayer Error List",layerLoadErrorMessages.join("<br>"));}\n'
        lines += "setTimeout(function(){if(layerLoadErrorMessages.length===0){$('#close-modal-button').click();}}, 2500);\n"

        #Iterate across each map command
        for mapCommand in self.mapCommandList:
            lines += mapCommand + '\n'
        
        # Set location of query outputs
        lines += 'queryWindowMode = "{}"\n'.format(self.queryWindowMode)
        
        lines+= "}"
        
        #Write out js file
        self.ee_run = os.path.join(ee_run_dir, '{}.js'.format(self.ee_run_name))
        oo = open(self.ee_run,'w')
        oo.writelines(lines)
        oo.close()
        # time.sleep(5)

        # if not self.isNotebook:
        #     self.Map.save(os.path.join(folium_html_folder,folium_html))
        #     if not geeView.isPortActive(self.port):
        #         print('Starting local web server at: http://localhost:{}/{}/'.format(self.port,geeView.geeViewFolder))
        #         geeView.run_local_server(self.port)
        #         print('Done')
        #     else:
        #         print('Local web server at: http://localhost:{}/{}/ already serving.'.format(self.port,geeView.geeViewFolder))
        #     if open_browser:
        #         geeView.webbrowser.open('http://localhost:{}/{}/{}'.format(self.port,geeView.geeViewFolder,folium_html),new = 1)

        # else:
        #     display(self.Map)
        
        if not isPortActive(self.port):
            print('Starting local web server at: http://localhost:{}/{}/'.format(self.port,geeViewFolder))
            run_local_server(self.port)
            print('Done')
            

        else:
            print('Local web server at: http://localhost:{}/{}/ already serving.'.format(self.port,geeViewFolder))
            # print('Refresh browser instance')
       

        print('cwd',os.getcwd())
        if IS_COLAB:
            proxy_js = "google.colab.kernel.proxyPort({})".format(self.port)
            proxy_url = eval_js(proxy_js)
            geeView_proxy_url = '{}geeView/?projectID={}&accessToken={}'.format(proxy_url,self.project,self.accessToken)
            print('Colab Proxy URL:',geeView_proxy_url)
            viewerFrame = IFrame(src=geeView_proxy_url, width='100%', height='{}px'.format(iframe_height))
            display(viewerFrame)
        elif IS_WORKBENCH:
            if self.proxy_url == None:
                self.proxy_url = input('Please enter current URL Workbench Notebook is running from (e.g. https://code-dot-region.notebooks.googleusercontent.com/): ')
            self.proxy_url = baseDomain(self.proxy_url)
            geeView_proxy_url = '{}/proxy/{}/geeView/?projectID={}&accessToken={}'.format(self.proxy_url,self.port,self.project,self.accessToken)
            print('Workbench Proxy URL:',geeView_proxy_url)
            viewerFrame = IFrame(src=geeView_proxy_url, width='100%', height='{}px'.format(iframe_height))
            display(viewerFrame)
        else:
            url = 'http://localhost:{}/{}/?projectID={}&accessToken={}'.format(self.port,geeViewFolder,self.project,self.accessToken)
            print('geeView URL:',url)
            if not self.isNotebook or open_browser:
                webbrowser.open(url,new = 1)
            elif open_browser == False and open_iframe:
                self.IFrame = IFrame(src=url, width='100%', height='{}px'.format(iframe_height))
            else:
                self.IFrame = IFrame(src=url, width='100%', height='{}px'.format(iframe_height))
                display(self.IFrame)
    def clearMap(self):
        self.layerNumber = 1
        self.idDictList = []
        self.mapCommandList  = []
    def setMapTitle(self,title):
        title_command = f'Map.setTitle("{title}")'
        if title_command not in self.mapCommandList:
            self.mapCommandList.append(title_command)
    def setTitle(self,title):
        self.setMapTitle(title)

    # Functions to set various click query properties
    def setQueryCRS(self,crs):
        print('Setting click query crs to: {}'.format(crs))
        cmd = f'Map.setQueryCRS("{crs}");'
        if cmd not in self.mapCommandList:
            self.mapCommandList.append(cmd)
    def setQueryScale(self,scale):
        print('Setting click query scale to: {}'.format(scale))
        cmd = f'Map.setQueryScale({scale});'
        if cmd not in self.mapCommandList:
            self.mapCommandList.append(cmd)
    def setQueryTransform(self,transform):
        print('Setting click query transform to: {}'.format(transform))
        cmd = f'Map.setQueryTransform({transform});'
        if cmd not in self.mapCommandList:
            self.mapCommandList.append(cmd)
    def setQueryPrecision(self,chartPrecision = 3,chartDecimalProportion=0.25):
        print('Setting click query precision to: {}'.format(chartPrecision))
        cmd = f'Map.setQueryPrecision({chartPrecision},{chartDecimalProportion});'
        if cmd not in self.mapCommandList:
            self.mapCommandList.append(cmd)
    def setQueryDateFormat(self,defaultQueryDateFormat = 'YYYY-MM-dd'):
        print('Setting default query date format to: {}'.format(defaultQueryDateFormat))
        cmd = f'Map.setQueryDateFormat("{defaultQueryDateFormat}");'
        if cmd not in self.mapCommandList:
            self.mapCommandList.append(cmd)
    def setQueryBoxColor(self,color):
        print('Setting click query box color to: {}'.format(color))
        cmd = f'Map.setQueryBoxColor("{color}");'
        if cmd not in self.mapCommandList:
            self.mapCommandList.append(cmd)
    # Functions to handle location of query outputs
    def setQueryWindowMode(self,mode):
        self.queryWindowMode = mode
    def setQueryToInfoWindow(self):
        self.setQueryWindowMode('infoWindow')
    def setQueryToSidePane(self):
        self.setQueryWindowMode('sidePane')
    
    # Turn on query inspector 
    def turnOnInspector(self):
        # self.mapCommandList.append("$('#tools-collapse-div').addClass('show')")
        query_command = "Map.turnOnInspector();"
        if query_command not in self.mapCommandList:
            self.mapCommandList.append(query_command)

    # Functions to handle setting query output y labels
    def setYLabelMaxLength(self,maxLength):
        command = f"yLabelMaxLength = {maxLength}"
        if command not in self.mapCommandList:
            self.mapCommandList.append(command)
    def setYLabelBreakLength(self,maxLength):
        command = f"yLabelBreakLength = {maxLength}"
        if command not in self.mapCommandList:
            self.mapCommandList.append(command)
    def setYLabelMaxLines(self,maxLines):
        command = f"yLabelMaxLines = {maxLines}"
        if command not in self.mapCommandList:
            self.mapCommandList.append(command)
    def setYLabelFontSize(self,fontSize):
        command = f"yLabelFontSize = {fontSize}"
        if command not in self.mapCommandList:
            self.mapCommandList.append(command)

    # Specify whether layers can be re-ordered by the user
    def setCanReorderLayers(self,canReorderLayers):
        command = f"Map.canReorderLayers = {str(canReorderLayers).lower()};"
        if command not in self.mapCommandList:
            self.mapCommandList.append(command)
    # Functions to handle batch layer toggling       
    def turnOffAllLayers(self):
        update = {'visible':'false'}
        self.idDictList = [{**d,**update} for d in self.idDictList]
    def turnOnAllLayers(self):
        update = {'visible':'true'}
        self.idDictList = [{**d,**update} for d in self.idDictList]
#Instantiate Map object
Map = mapper()


# print(ee.data._initialized  )
# ee.Authenticate(auth_mode='localhost')
# ee.Authenticate(force=True)

# ee.Authenticate()
# ee.Initialize()#project='lcms-292214')
# ee.data.setCloudApiUserProject('rcr-gee')

# # ee.Initialize()
# print(ee.data._cloud_api_user_project  )
# print(refreshToken())
# f = ee.FeatureCollection('projects/rcr-gee-2/assets/test/DAMAGE_AREAS_FLAT_AllYears_CONUS_Rgn9')
# print(ee.Image().getInfo())
# Map.addLayer(f,{'palette':'F00'})
# Map.turnOnInspector()





"""
   Copyright 2024 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

# This script provides a view of LCMAP and LCMS GEE collections to help understand the strengths of each
# It displays the land cover and land use products as well as change products from each program
####################################################################################################
import os,sys
sys.path.append(os.getcwd())

#Module imports
import geeViz.getImagesLib as getImagesLib
import geeViz.changeDetectionLib as changeDetectionLib
ee = getImagesLib.ee
Map = getImagesLib.Map
Map.clearMap()
####################################################################################################
# User Parameters

# Define the early time period 
preStartYear = 1985
preEndYear = 1990

# Define the more recent time period
postStartYear = 2020
postEndYear = 2022

# Specify whether to add time lapses of products. If True, loading the viewer will take much much much much longer
addTimelapses = False
####################################################################################################
# Bring in LCMAP outputs
# This part was adapted from info available in the awesome-gee-community-datasets
# https://samapriya.github.io/awesome-gee-community-datasets/projects/lcmap/
# Playground example: https://code.earthengine.google.com/791aa894ce0abfe1a9eb1dc478bbc5d7
# Their outputs are divided into LC (land cover) and SC (spectral change)
# More details about their different products can be found here:
# https://www.sciencedirect.com/science/article/pii/S003442571930375X
# LCMAP methods can be found here: https://www.sciencedirect.com/science/article/pii/S003442571930375X
# LCMAP data can be downloaded from: https://www.usgs.gov/core-science-systems/eros/lcmap/lcmap-data-access
# LCMAP data can also be viewed here: https://eros.usgs.gov/lcmap/viewer/
lcachg = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/LCACHG")
lcpconf = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/LCPCONF")
lcpri = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/LCPRI")
lcsconf = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/LCSCONF")
lcsec = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/LCSEC")
sclast = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/SCLAST")
scmag = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/SCMAG")
scmqa = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/SCMQA")
scstab = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/SCSTAB")
sctime = ee.ImageCollection("projects/sat-io/open-datasets/LCMAP/SCTIME")


# Manually pull in color palettes from Gena's color palette json
# palettes = require('users/gena/packages:palettes')
# Available here: https://github.com/gee-community/ee-palettes
lcpri_palette = ['E60000','A87000','E3E3C2','1D6330','476BA1','BAD9EB','FFFFFF','B3B0A3','A201FF']
lc_names = ['Developed','Cropland','Grass/Shrub','Tree Cover','Water','Wetlands','Ice/Snow','Barren','Class Change']

# Set up the LCMAP land cover legend and lookup tables
lc_legend_dict = {}
lc_lookup_dict = {}
for i in range(0,len(lc_names )):
  lc_legend_dict[str(i+1) + '- '+lc_names[i]] =lcpri_palette[i] 
  lc_lookup_dict[i+1] = str(i+1) + '- '+lc_names[i]

#Other palettes found in the Playground example script
# lcsec_palette = ['E60000','A87000','E3E3C2','1D6330','476BA1','BAD9EB','FFFFFF','B3B0A3']
lcachg_palette = ['E60000','A87000','E3E3C2','1D6330','476BA1','BAD9EB','FFFFFF','B3B0A3','A201FF']
# sclast_palette = ['FFC7AA','F87E45','CC764E','86A7B6','46A4EE','7954C8','7A24AA','432172']
# scstab_palette = ['BA4E16','EE964D','FFE29C','F4FBC1','E1F3C3','BCE6CA','46989C']
# scmqa_palette = ['000000','A900E6','DF73FF','F5F5E3','DB8A00','924900','9C9C9C','FFFFFF']

# Set up some visualization dictionaries
lc_viz = {'reducer':ee.Reducer.mode(),'min':1,'max':9,'palette':lcpri_palette,'classLegendDict':lc_legend_dict,'queryDict':lc_lookup_dict}
loss_viz = {'min':1985,'max':2022,'palette':changeDetectionLib.lossYearPalette}
gain_viz = {'min':1985,'max':2022,'palette':changeDetectionLib.gainYearPalette}
change_viz = {'min':1985,'max':2022,'palette':['00F','F0F']}

# Map.addLayer(sctime.max(),{'min':100,'max':292,'palette':['151d44', '156c72', '7eb390', 'fdf5f4', 'db8d77', '9c3060', '340d35']},'SCTIME',True)
# Map.addLayer(scmag.max(),{'min':651,'max':3700,'palette':['d7f9d0', 'a2d595', '64b463', '129450', '126e45', '1a482f', '122414']},'SCMAG',True)
# Map.addLayer(ee.Image(sclast.sort('system:time_start',false).first()).mask(ee.Image(sclast.sort('system:time_start',false).first()).gt(0)),{min:518,max:4600,palette:sclast_palette},'SCLAST',false)
# Map.addLayer(scstab.sort('system:time_start',false).first(),{min:70,max:13000,palette:scstab_palette},'SCSTAB',false)
# Map.addLayer(scmqa.sort('system:time_start',false).first().remap([0,4,6,8,14,24,44,54],[0,1,2,3,4,5,6,7]).mask(ee.Image(scmqa.sort('system:time_start',false).first()).gt(0)),{min:0,max:7,palette:scmqa_palette},'SCMQA',false)
####################################################################################################
# Bring in LCMS GEE collection over CONUS
# LCMS' homepage can be found here: https://data.fs.usda.gov/geodata/rastergateway/LCMS/index.php
# LCMS methods are described here: https://data.fs.usda.gov/geodata/rastergateway/LCMS/LCMS_v2021-7_Methods.pdf
# LCMS data can also be viewed and downloaded here: https://apps.fs.usda.gov/lcms-viewer
# LCMS GEE data collections are available at:
  # https://developers.google.com/earth-engine/datasets/catalog/USFS_GTAC_LCMS_v20201-7 (CONUS and Southeastern AK)
  # https://developers.google.com/earth-engine/datasets/catalog/USFS_GTAC_LCMS_v2020-6 (Puerto Rico and US Virgin Islands
# An in-depth look at the model predictor variables that go into making LCMS maps can be found here:https://apps.fs.usda.gov/lcms-viewer/lcms-base-learner.html
lcms = ee.ImageCollection("USFS/GTAC/LCMS/v2022-8").filter(ee.Filter.eq('study_area','CONUS'))
####################################################################################################

####################################################################################################
#This section adds the land cover and land use maps from LCMAP and LCMS

# Pull the LCMAP pre and post land cover data
lcmap_pre = lcpri.filter(ee.Filter.calendarRange(preStartYear,preEndYear,'year'))
lcmap_post = lcpri.filter(ee.Filter.calendarRange(postStartYear,postEndYear,'year'))

# Pull the LCMS pre and post land cover data
lcms_pre = lcms.filter(ee.Filter.calendarRange(preStartYear,preEndYear,'year'))
lcms_post = lcms.filter(ee.Filter.calendarRange(postStartYear,postEndYear,'year'))



# Add time lapses of LCMAP and LCMS land cover products if specified
if addTimelapses:
  Map.addTimeLapse(lcachg,lc_viz,'LCMAP LC Change')
  Map.addTimeLapse(lcpri,lc_viz,'LCMAP Primary Land Cover')
  Map.addTimeLapse(lcsec,lc_viz,'LCMAP Secondary Land Cover')
  Map.addTimeLapse(lcms.select(['Land_Cover']),{'autoViz':True},'LCMS Land Cover')
  Map.addTimeLapse(lcms.select(['Land_Use']),{'autoViz':True},'LCMS Land Use')

# Add the early and recent land cover and land use mode maps
# LCMAP's LC outputs combine land cover and land use, so cross-walking these non-mutually exclusive classes to those of LCMS can pose challenges
# However, take note how well LCMAP maps agriculture and developed areas - it is much cleaner and seems much more accurate than LCMS.
# LCMAP does about as well as LCMS at mapping trees. They both do fairly well in sparse tree cover areas.
# Since wetlands are not exclusive of any land cover or land use, it is difficult to tell what land cover is over areas classified by LCMAP as wetland
# This is also true of LCMS' non forest wetland land use class - there are areas of rangeland and agriculture that could fall into non forest wetland
# One area the LCMS land cover maps do better is with water. 
# LCMAP LC outputs tend to not change the water extent of fluctuating waterbodies very quickly
Map.addLayer(lcmap_pre,lc_viz, 'LCMAP LC {}-{} mode'.format(preStartYear,preEndYear),False)
Map.addLayer(lcmap_post,lc_viz, 'LCMAP LC {}-{} mode'.format(postStartYear,postEndYear),False)

for t in ['Cover','Use']:
  Map.addLayer(lcms_pre.select(['Land_{}'.format(t)]),{'reducer':ee.Reducer.mode(),'autoViz':True}, 'LCMS L{} {}-{} mode'.format(t[0],preStartYear,preEndYear),False)
  Map.addLayer(lcms_post.select(['Land_{}'.format(t)]),{'reducer':ee.Reducer.mode(),'autoViz':True}, 'LCMS L{} {}-{} mode'.format(t[0],postStartYear,postEndYear),False)

####################################################################################################
# This section adds the change maps from LCMAP and LCMS

# Function for getting a precise change break date from CCDC-based outputs
def getYrMskPrecise(img):
  yr = ee.Date(img.get('system:time_start')).get('year')
  return ee.Image(yr).add(img.divide(365)).float().updateMask(img.mask())

# Function for getting a more general integer year date of change
def getYrMsk(img):
  yr = ee.Date(img.get('system:time_start')).get('year')
  return ee.Image(yr).int16().updateMask(img.mask())

# Find the most recent LCMAP spectral change date (YYYY.dd where .dd is the fraction of the year the break occurred)
lcmap_change_yr  = sctime.map(getYrMskPrecise).max()

# Pull apart LCMS fast and slow loss and find the most recent year of each 
lcms_fast_loss_yr = lcms.select(['Change']).map(lambda img:getYrMsk(img.updateMask(img.eq(3)))).max()
lcms_slow_loss_yr = lcms.select(['Change']).map(lambda img:getYrMsk(img.updateMask(img.eq(2)))).max()
lcms_gain_yr = lcms.select(['Change']).map(lambda img:getYrMsk(img.updateMask(img.eq(4)))).max()


# As of version 2020.5, LCMS produces vegetation cover slow loss, fast loss, and gain change outputs. 
# LCMAP version 1.0 produces several change outputs based on the spectral change detected by CCDC
# The most analagous change product is the spectral change time (SCTIME)
Map.addLayer(lcms_fast_loss_yr,loss_viz,'LCMS Most Recent Fast Loss Year',False)
Map.addLayer(lcms_slow_loss_yr,loss_viz,'LCMS Most Recent Slow Loss Year',False)
Map.addLayer(lcms_gain_yr,gain_viz,'LCMS Most Recent Gain Year',False)

Map.addLayer(lcmap_change_yr,change_viz,'LCMAP Most Recent SC Date',True)

####################################################################################################
# Since LCMAP is largely based on CCDC, understanding how the delivered change outputs relate to raw CCDC outputs helps 
# identify the strengths and weaknesses of the approach

# Bring in the LCMS CCDC output that is similar to what LCMAP uses
ccdcImg = ee.ImageCollection('projects/lcms-292214/assets/CONUS-LCMS/Base-Learners/CCDC-Collection-1984-2022')\
          .select(['tStart','tEnd','tBreak','changeProb','red.*','nir.*','swir1.*','swir2.*','NDVI.*']).mosaic()

# Pull out the most recent date of change
changeObj = changeDetectionLib.ccdcChangeDetection(ccdcImg,'NDVI');
Map.addLayer(ccdcImg,{'opacity':0},'RAW CCDC Output',False)
Map.addLayer(changeObj['mostRecent']['loss']['year'],loss_viz,'LCMS CCDC Most Recent Loss Year')
Map.addLayer(changeObj['mostRecent']['gain']['year'],gain_viz,'LCMS CCDC Most Recent Gain Year')


# This produces a chart of the harmonic models from CCDC and the breaks to help further understand how the outputs are created
# Apply the CCDC harmonic model across a time series
# First get a time series of time images 
yearImages = changeDetectionLib.getTimeImageCollection(1984,2022,1,365,0.1)

#Then predict the CCDC models
fitted = changeDetectionLib.predictCCDC(ccdcImg,yearImages,False,[1,2,3])

Map.addLayer(fitted.select(['NDVI_CCDC_fitted']),{'opacity':0},'Fitted CCDC NDVI',True);
####################################################################################################
Map.setTitle('LCMAP LCMS Viewer')
Map.turnOnInspector()
Map.setQueryDateFormat('YYYY')
Map.view()


"""
   Copyright 2024 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
#Script to help with basic cloudStorage management
#Intended to work within the geeViz package
######################################################################
import geeViz.geeView
from google.cloud import storage
######################################################################
# Function to list all blobs in a bucket
# Returns a list of blob objects
# Use the blob.name to get the name
def list_blobs(bucket_name):
    #storage client instance
    storage_client = storage.Client(project = geeViz.geeView.project_id)

    #get bucket by name
    bucket = storage_client.get_bucket(bucket_name)

    #list files stored in bucket
    all_blobs = bucket.list_blobs()
    return list(all_blobs)
######################################################################
# Function to batch rename blobs in a bucket
def rename_blobs(bucket_name,old_name,new_name):
    """Renames a group of blobs."""
    #storage client instance
    storage_client = storage.Client(project = geeViz.geeView.project_id)

    #get bucket by name
    bucket = storage_client.get_bucket(bucket_name)

    #list files stored in bucket
    all_blobs = list(bucket.list_blobs())

    #Renaming all files to lowercase:
    for blob in all_blobs:
        on = str(blob.name)
        nn = on.replace(old_name,new_name)
        new_blob = bucket.rename_blob(blob, new_name= nn)
        print('Blob {} has been renamed to {}'.format( on,nn))
######################################################################
# Return wether a filename exists or not
def gcs_exists(bucket,filename):
    storage_client = storage.Client(project = geeViz.geeView.project_id)
    stats = storage.Blob(bucket=storage_client.bucket(bucket), name=filename).exists(storage_client)
    return stats
######################################################################
# !! Dangerous !! - cannot be undone
# Delete a specified filename
def delete_blob(bucket,filename):
    storage_client = storage.Client(project = geeViz.geeView.project_id)
    out = storage.Blob(bucket=storage_client.bucket(bucket), name=filename).delete(storage_client)
    print('Deleted:',filename)
######################################################################




"""
   Copyright 2024 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

# Example of how to use a local shapefile in geeViz workflows
# This uses geemap to convert the shapefile to geojson and then an ee object 
####################################################################################################
import geemap
import geeViz.getImagesLib as gil
Map = gil.Map
ee = gil.ee

# Specify dates to make an image composite
startYear = 2023
endYear = 2023
startJulian = 150
endJulian = 250

# Provide a shapefile
shp = gil.os.path.join(gil.sys.path[0],"data/gadm41_CHE_shp/gadm41_CHE_0.shp")

# Convert it to an ee object
studyArea = geemap.shp_to_ee(shp)

# Get some images (use the bounds to avoid memory errors)
s2s = gil.getProcessedSentinel2Scenes(studyArea.geometry().bounds(),startYear,endYear,startJulian,endJulian)

# Add the composite and study area to the map
Map.addLayer(s2s.median(),gil.vizParamsFalse,'Composite')


Map.addLayer(studyArea,{'strokeColor':'0FF','strokeWeight':3},'Study Area')

Map.centerObject(studyArea)
Map.turnOnInspector()
Map.view()